import json
import sys
import numpy

from material.orthotropic import Orthotropic
from ply.ply import Ply
from laminate.laminate import Laminate
from panel.panel import Panel
from loads.loads import Loads
from lambend.lambend import Lambend
from lambend.__version__ import __version__
numpy.set_printoptions(threshold=sys.maxsize, linewidth=1000)

class LambendIO(object):

    def __init__(self, materialFile=None, plyFile=None, laminateFile=None,
    panelFile=None, loadsFile=None, outputFile=None, number_of_terms=None,
    number_of_points_x=None, number_of_points_y=None,
    output_K="no", output_P="no"):
        """Initializes LambendIO instance

        Parameters
        ----------

        materialFile : str

            Full path filename of material json file (required).

        plyfile : str

            Full path filename of ply json file (required).

        laminateFile : str

            Full path filename of laminate json file (required).

        panelFile : str

            Full path filename of panel json file (required).

        loadsFile : str

            Full path filename of loads json file (required).

        outputFile : str

            Full path filename of output file (required).

        number_of_terms : int

            Number of terms to be used in analysis (required).

        number_of_points_x : int

            Number of points in x direction where outputs are reported (required).

        number_of_points_y : int

            Number of points in y direction where outputs are reported (required).
       
        """

        self.materialFile = materialFile
        self.plyFile = plyFile
        self.laminateFile = laminateFile
        self.panelFile = panelFile
        self.loadsFile = loadsFile
        self.outputFile = outputFile
        self.number_of_terms = number_of_terms
        self.number_of_points_x = number_of_points_x
        self.number_of_points_y = number_of_points_y
        self.output_K = output_K
        self.output_P = output_P

        self._invalidate()

    def _invalidate(self):
        self._materials = {}
        self._plies = {}
        self._laminates = {}
        self._panels = {}
        self._loads = {}
        self._results = []
        self._pairs = []
        self._unpairs = []

    @property
    def materials(self):
        """Returns dictionary of materials read from input file. The key 
        in the dictionary is the material id.

        Returns
        -------

        materialDict : dict

            {Material.matID : Material}
        
        """

        if not self._materials:
            with open(self.materialFile, 'r') as fh:
                jsonData = json.load(fh)
            for materialDict in jsonData["materials"]:
                matID = materialDict["matID"]
                E1 = materialDict["E1"]
                E2 = materialDict["E2"]
                G12 = materialDict["G12"]
                G13 = materialDict["G13"]
                G23 = materialDict["G23"]
                v12 = materialDict["v12"]
                alpha1 = materialDict["alpha1"]
                alpha2 = materialDict["alpha2"]
                beta1 = materialDict["beta1"]
                beta2 = materialDict["beta2"]
                self._materials[matID] = Orthotropic(matID=matID, E1=E1, E2=E2, 
                G12=G12, G13=G13, G23=G23, v12=v12, alpha1=alpha1,
                alpha2=alpha2, beta1=beta1, beta2=beta2)

        return self._materials

    @property
    def plies(self):
        """Returns dictionary of plies read from input file. The key 
        in the dictionary is the ply id.

        Returns
        -------

        plyDict : dict

            {Ply.plyID : Ply, ...}
        
        """

        if not self._plies:
            with open(self.plyFile, 'r') as fh:
                jsonData = json.load(fh)
            for plyDict in jsonData["plies"]:
                plyID = plyDict["plyID"]
                angle = plyDict["angle"]
                material = self.materials[plyDict["material"]]
                thickness = plyDict["thickness"]
                self._plies[plyID] = Ply(plyID=plyID, angle=angle, 
                material=material, thickness=thickness)

        return self._plies
    
    @property
    def laminates(self):
        """Returns dictionary of laminates read from input file. The key 
        in the dictionary is the laminate id.

        Returns
        -------

        laminateDict : dict

            {Laminate.lamID : Laminate, ...}
        
        """
        if not self._laminates:
            with open(self.laminateFile, 'r') as fh:
                jsonData = json.load(fh)
            for laminateDict in jsonData["laminates"]:
                lamID = laminateDict["lamID"]
                stacking_sequence = [self.plies[str(plyID)] for plyID in 
                laminateDict["stacking_sequence"]]
                offset = laminateDict["offset"]
                self._laminates[lamID] = Laminate(lamID=lamID, 
                stacking_sequence=stacking_sequence, offset=offset)

                # Read A, B, D from input file directly
                if "A" in laminateDict:
                    A11, A12, A16, A22, A26, A66, A44, A45, A55 = laminateDict["A"]
                    self._laminates[lamID]._A = numpy.array([[A11, A12, A16], [A12, A22, A26], [A16, A26, A66]])
                    self._laminates[lamID]._A44 = A44
                    self._laminates[lamID]._A45 = A45
                    self._laminates[lamID]._A55 = A55
                if "B" in laminateDict:
                    B11, B12, B16, B22, B26, B66 = laminateDict["B"]
                    self._laminates[lamID]._B = numpy.array([[B11, B12, B16], [B12, B22, B26], [B16, B26, B66]])
                if "D" in laminateDict:
                    D11, D12, D16, D22, D26, D66 = laminateDict["D"]
                    self._laminates[lamID]._D = numpy.array([[D11, D12, D16], [D12, D22, D26], [D16, D26, D66]])

        return self._laminates

    @property
    def panels(self):
        """Returns dictionary of panels read from input file. The key 
        in the dictionary is the panel id.

        Returns
        -------

        panelDict : dict

            {Panel.panelID : Panel, ...}
        
        """
        if not self._panels:
            with open(self.panelFile, 'r') as fh:
                jsonData = json.load(fh)
            for panelDict in jsonData["panels"]:
                panelID = panelDict["panelID"]
                length = panelDict["length"]
                width = panelDict["width"]
                laminate = self.laminates[panelDict["laminate"]]
                boundary_condition = panelDict["boundary_condition"]
                self._panels[panelID] = Panel(panelID=panelID, length=length, width=width,
                laminate=laminate, boundary_condition=boundary_condition)
                self._panels[panelID].number_of_terms = self.number_of_terms

        return self._panels

    @property
    def loads(self):
        """Returns dictionary of loads read from input file. The key
        in the dictionary is the loadscase or loads id
        
        Returns
        -------
        
        loadsDict : dict

            {Loads.loadsID_Panel.panelID : Loads, ...}
            
        """

        if not self._loads:
            with open(self.loadsFile, 'r') as fh:
                jsonData = json.load(fh)
                for loadsDict in jsonData["loads"]:
                    loadsID = loadsDict["loadsID"]
                    panelID = loadsDict["panelID"]
                    loadsType = loadsDict["type"]
                    try:
                        p0 = loadsDict["p0"]
                    except:
                        p0 = 0.0

                    try:
                        P = loadsDict["P"]
                    except:
                        P = 0.0
                    try:
                        xCG = loadsDict["xCG"]
                    except:
                        xCG = 0.0
                    try:
                        yCG = loadsDict["yCG"]
                    except:
                        yCG = 0.0
                    try:
                        xLength = loadsDict["xLength"]
                    except:
                        xLength = 0.0
                    try:
                        yLength = loadsDict["yLength"]
                    except:
                        yLength = 0.0
                    try:
                        Nxx = loadsDict["Nxx"]
                    except:
                        Nxx = 0.0
                    try:
                        Nyy = loadsDict["Nyy"]
                    except:
                        Nyy = 0.0
                    try:
                        Nxy = loadsDict["Nxy"]
                    except:
                        Nxy = 0.0
                    try:
                        deltaT = loadsDict["deltaT"]
                    except:
                        deltaT = 0.0
                    try:
                        deltaC = loadsDict["deltaC"]
                    except:
                        deltaC = 0.0
                    self._loads[loadsID+"_"+panelID] = Loads(loadsID=loadsID, panelID=panelID,
                    p0=p0, P=P, xCG=xCG, yCG=yCG, xLength=xLength, yLength=yLength,
                    loadsType = loadsType,
                    Nxx=Nxx, Nyy=Nyy, Nxy=Nxy, deltaT=deltaT, deltaC=deltaC)

        return self._loads

    @property
    def pairs(self):
        """Returns a list of analysis pairs of (Panel, Loads).
        
        Returns
        -------
        
        pairs : list
            [(Panel, Loads), ...]
            
        """

        if not self._pairs:
            for k, lds in self.loads.items():
                pnl = self.panels[lds.panelID]
                self._pairs.append((pnl, lds))

        
        return self._pairs

    @property
    def results(self):
        """Returns a list of Lambend for each pair of (panel, loads)
        
        Returns
        -------

        solver001List : list
            [Lambend, ...]

        """
        if not self._results:
            for pnl, lds in self.pairs:
                lmb = Lambend(panel=pnl, loads=lds, 
                number_of_terms=self.number_of_terms,
                number_of_points_x=self.number_of_points_x,
                number_of_points_y=self.number_of_points_y)
                self._results.append(lmb)
        return self._results

    def _writeResults(self):
        """Writes output in outputfile"""

        outputData = "3pc-lambend, v"+__version__+"\n\n"
        KData = ""
        PData = ""
        for lmb in self.results:
            postedMat = []
            postedMatTH = []
            outputData +=  "LOADS ID  PANEL ID  \n"
            outputData += "{:<10s}".format(lmb.loads.loadsID)
            outputData += "{:<10s}\n\n".format(lmb.panel.panelID)

            outputData += "PANEL GEOMETRY\n"
            outputData += "LENGTH: {:<14.6e}\n".format(lmb.panel.length)
            outputData += "WIDTH : {:<14.6e}\n\n".format(lmb.panel.width)

            outputData += "LOADS DESCRIPTION\n"
            outputData += "TYPE: {:<20s}\n".format(lmb.loads.type.upper())
            if lmb.loads.type == "uniform pressure" or lmb.loads.type == "sinusoidal pressure" or lmb.loads.type == "hydrostatic pressure in x direction" or lmb.loads.type == "hydrostatic pressure in y direction" or lmb.loads.type == "triangular pressure in x direction" or lmb.loads.type == "triangular pressure in y direction":
                outputData += "p0: {:<14.6e}\n".format(lmb.loads.p0)
            elif lmb.loads.type == "partial pressure":
                outputData += "p0: {:<14.6e}\n".format(lmb.loads.p0)
                outputData += "xCG: {:<14.6e}\n".format(lmb.loads.xCG)
                outputData += "yCG: {:<14.6e}\n".format(lmb.loads.yCG)
                outputData += "xLength: {:<14.6e}\n".format(lmb.loads.xLength)
                outputData += "yLength: {:<14.6e}\n".format(lmb.loads.yLength)
            elif lmb.loads.type == "point load":
                outputData += "P: {:<14.6e}\n".format(lmb.loads.P)
                outputData += "xCG: {:<14.6e}\n".format(lmb.loads.xCG)
                outputData += "yCG: {:<14.6e}\n".format(lmb.loads.yCG)
            outputData += "\n"

            outputData += "ANALYSIS OPTIONS\n"
            outputData += "m = {:<14d}\n".format(lmb.number_of_terms)
            outputData += "n = {:<14d}\n\n".format(lmb.number_of_terms)

            outputData += "OUTPUT OPTIONS\n"
            outputData += "NUMBER OF POINTS IN X DIR: {:<14d}\n".format(lmb.number_of_points_x)
            outputData += "NUMBER OF POINTS IN Y DIR: {:<14d}\n\n".format(lmb.number_of_points_y)

            outputData += "{:^87s}\n".format("MATERIAL PROPERTIES")
            outputData += "ID            E1            E2            G12           G23           G13           v12           \n"
            for ply in lmb.panel.laminate.stacking_sequence:
                if ply.material.matID in postedMat:
                    continue
                postedMat.append(ply.material.matID)
                if len(ply.material.matID) <= 13:
                    outputData += "{:<14s}".format(ply.material.matID)
                else:
                    outputData += "{:<14s}".format(ply.material.matID[0:12])
                outputData += "{:<14.6e}".format(ply.material.E1)
                outputData += "{:<14.6e}".format(ply.material.E2)
                outputData += "{:<14.6e}".format(ply.material.G12)
                outputData += "{:<14.6e}".format(ply.material.G23)
                outputData += "{:<14.6e}".format(ply.material.G13)
                outputData += "{:<14.6e}\n".format(ply.material.v12)
            outputData += "\n"

            outputData += "{:^61s}\n".format("HYGROTHERMAL MATERIAL PROPERTIES")
            outputData += "ID            alpha1        alpha2        beta1         beta2         \n"
            for ply in lmb.panel.laminate.stacking_sequence:
                if ply.material.matID in postedMatTH:
                    continue
                postedMatTH.append(ply.material.matID)
                if len(ply.material.matID) <= 13:
                    outputData += "{:<14s}".format(ply.material.matID)
                else:
                    outputData += "{:<14s}".format(ply.material.matID[0:12])
                outputData += "{:<14.6e}".format(ply.material.alpha1)
                outputData += "{:<14.6e}".format(ply.material.alpha2)
                outputData += "{:<14.6e}".format(ply.material.beta1)
                outputData += "{:<14.6e}\n".format(ply.material.beta2)
            outputData += "\n"

            outputData += "{:^28s}\n".format("LAMINATE GEOMETRY")
            outputData += "STACKING SEQUENCE (PLY ANG): ["
            for i, ply in enumerate(lmb.panel.laminate.stacking_sequence, start=1):
                if i < len(lmb.panel.laminate.stacking_sequence):
                    outputData += "{:<+14.1f}, ".format(ply.angle)
                else:
                    outputData += "{:<+14.1f}]\n".format(ply.angle)
            
            outputData += "STACKING SEQUENCE (PLY MAT): ["
            for i, ply in enumerate(lmb.panel.laminate.stacking_sequence, start=1):
                if i < len(lmb.panel.laminate.stacking_sequence):
                    if len(ply.material.matID) <= 13:
                        outputData += "{:<14s}, ".format(ply.material.matID)
                    else:
                        outputData += "{:<14s}, ".format(ply.material.matID[0:12])
                else:
                    if len(ply.material.matID) <= 13:
                        outputData += "{:<14s}]\n".format(ply.material.matID)
                    else:
                        outputData += "{:<14s}]\n".format(ply.material.matID[0:12])

            outputData += "TOTAL THICKNESS: {:<14.6e}\n".format(lmb.panel.laminate.thickness)
            outputData += "OFFSET: {:<14.6e}\n".format(lmb.panel.laminate.offset)
            outputData += "\n"

            outputData += "{:^42s}\n".format("LAMINATE PROPERTIES")
            outputData += "{:^42s}\n".format("A MATRIX")
            outputData += "{:<+14.6e}{:<+14.6e}{:<+14.6e}\n".format(lmb.panel.laminate.A[0,0], 
            lmb.panel.laminate.A[0,1], lmb.panel.laminate.A[0,2])
            outputData += "{:<+14.6e}{:<+14.6e}{:<+14.6e}\n".format(lmb.panel.laminate.A[1,0], 
            lmb.panel.laminate.A[1,1], lmb.panel.laminate.A[1,2])
            outputData += "{:<+14.6e}{:<+14.6e}{:<+14.6e}\n\n".format(lmb.panel.laminate.A[2,0], 
            lmb.panel.laminate.A[2,1], lmb.panel.laminate.A[2,2])

            outputData += "{:^42s}\n".format("A MATRIX - TRANSVERSE SHEAR")
            outputData += "{:<+14.6e}{:<+14.6e}\n".format(lmb.panel.laminate.A44, 
            lmb.panel.laminate.A45)
            outputData += "{:<+14.6e}{:<+14.6e}\n\n".format(lmb.panel.laminate.A45, 
            lmb.panel.laminate.A55)

            outputData += "{:^42s}\n".format("B MATRIX")
            outputData += "{:<+14.6e}{:<+14.6e}{:<+14.6e}\n".format(lmb.panel.laminate.B[0,0], lmb.panel.laminate.B[0,1], 
            lmb.panel.laminate.B[0,2])
            outputData += "{:<+14.6e}{:<+14.6e}{:<+14.6e}\n".format(lmb.panel.laminate.B[1,0], lmb.panel.laminate.B[1,1], 
            lmb.panel.laminate.B[1,2])
            outputData += "{:<+14.6e}{:<+14.6e}{:<+14.6e}\n\n".format(lmb.panel.laminate.B[2,0], lmb.panel.laminate.B[2,1], 
            lmb.panel.laminate.B[2,2])

            if lmb.panel.laminate.offsetKey.upper() != "MIDDLE":
                outputData += "{:^42s}\n".format("B MATRIX - NO OFFSET")
                outputData += "{:<+14.6e}{:<+14.6e}{:<+14.6e}\n".format(lmb.panel.laminate.Bmid[0,0], lmb.panel.laminate.Bmid[0,1], 
                lmb.panel.laminate.Bmid[0,2])
                outputData += "{:<+14.6e}{:<+14.6e}{:<+14.6e}\n".format(lmb.panel.laminate.Bmid[1,0], lmb.panel.laminate.Bmid[1,1], 
                lmb.panel.laminate.Bmid[1,2])
                outputData += "{:<+14.6e}{:<+14.6e}{:<+14.6e}\n\n".format(lmb.panel.laminate.Bmid[2,0], lmb.panel.laminate.Bmid[2,1], 
                lmb.panel.laminate.Bmid[2,2])

            outputData += "{:^42s}\n".format("D MATRIX")
            outputData += "{:<+14.6e}{:<+14.6e}{:<+14.6e}\n".format(lmb.panel.laminate.D[0,0], lmb.panel.laminate.D[0,1], 
            lmb.panel.laminate.D[0,2])
            outputData += "{:<+14.6e}{:<+14.6e}{:<+14.6e}\n".format(lmb.panel.laminate.D[1,0], lmb.panel.laminate.D[1,1], 
            lmb.panel.laminate.D[1,2])
            outputData += "{:<+14.6e}{:<+14.6e}{:<+14.6e}\n\n".format(lmb.panel.laminate.D[2,0], lmb.panel.laminate.D[2,1], 
            lmb.panel.laminate.D[2,2])

            if lmb.panel.laminate.offsetKey.upper() != "MIDDLE":
                outputData += "{:^42s}\n".format("D MATRIX - NO OFFSET")
                outputData += "{:<+14.6e}{:<+14.6e}{:<+14.6e}\n".format(lmb.panel.laminate.Dmid[0,0], lmb.panel.laminate.Dmid[0,1], 
                lmb.panel.laminate.Dmid[0,2])
                outputData += "{:<+14.6e}{:<+14.6e}{:<+14.6e}\n".format(lmb.panel.laminate.Dmid[1,0], lmb.panel.laminate.Dmid[1,1], 
                lmb.panel.laminate.Dmid[1,2])
                outputData += "{:<+14.6e}{:<+14.6e}{:<+14.6e}\n\n".format(lmb.panel.laminate.Dmid[2,0], lmb.panel.laminate.Dmid[2,1], 
                lmb.panel.laminate.Dmid[2,2])

            outputData += "{:^130s}\n".format("LAMINATE INPLANE AND FLEXURAL ENGINEERING CONSTANTS")
            outputData += "Ex            Ey            Gxy           vxy           vyx           "
            outputData += "Efx           Efy           Gfxy          vfxy          vfyx          \n"
            outputData += "{:<+14.6e}{:<+14.6e}{:<+14.6e}{:<+14.6e}{:<+14.6e}{:<+14.6e}{:<+14.6e}{:<+14.6e}{:<+14.6e}{:<+14.6e}\n\n".format(lmb.panel.laminate.Ex,
            lmb.panel.laminate.Ey, lmb.panel.laminate.Gxy, lmb.panel.laminate.vxy, lmb.panel.laminate.vyx,
            lmb.panel.laminate.Efx, lmb.panel.laminate.Efy, lmb.panel.laminate.Gfxy, lmb.panel.laminate.vfxy,
            lmb.panel.laminate.vfyx)

            outputData += "{:^76s}\n".format("EFFECTIVE LAMINATE HYGROTHERMAL ENGINEERING CONSTANTS")
            outputData += "alphax        alphay        alphaxy       betax         betay         betaxy        \n"
            outputData += "{:<+14.6e}{:<+14.6e}{:<+14.6e}{:<+14.6e}{:<+14.6e}{:<+14.6e}\n\n".format(lmb.panel.laminate.alphax, 
            lmb.panel.laminate.alphay, lmb.panel.laminate.alphaxy, lmb.panel.laminate.betax, lmb.panel.laminate.betay, lmb.panel.laminate.betaxy)

            outputData += "{:^31s}\n".format("MECHANICAL LOADS")
            outputData += "NX            NY            NXY           \n"
            outputData += "{:<+14.6e}{:<+14.6e}{:<+14.6e}\n\n".format(lmb.loads.Nxx, lmb.loads.Nyy, lmb.loads.Nxy)

            outputData += "{:^46s}\n".format("THERMAL LOADS")
            outputData += "DELTAT        NXT           NYT           NXYT          \n"
            outputData += "{:<+14.6e}{:<+14.6e}{:<+14.6e}{:<+14.6e}\n\n".format(lmb.loads.deltaT, lmb.NT[0], lmb.NT[1], lmb.NT[2])

            outputData += "{:^46s}\n".format("HYGRAL LOADS")
            outputData += "DELTAC        NXH           NYH           NXYH          \n"
            outputData += "{:<+14.6e}{:<+14.6e}{:<+14.6e}{:<+14.6e}\n\n".format(lmb.loads.deltaC, lmb.NC[0], lmb.NC[1], lmb.NC[2])

            outputData += "{:^199s}\n".format("GRID POINTS, DISPLACEMENTS, ROTATIONS, FORCES, AND MOMENTS")
            outputData += "X             Y             U             V             W             PHIX          PHIY          "
            outputData += "NXX           NYY           NXY           MXX           MYY           MXY           QYZ           QXZ           \n"
            for loc, uloc, vloc, wloc, phixloc, phiyloc, NMBloc, QBloc in zip(lmb.loc, lmb.uloc, lmb.vloc, lmb.wloc, lmb.phixloc, lmb.phiyloc, lmb.NMBloc, lmb.QBloc):
                outputData += "{:<14.6e}".format(loc[0])
                outputData += "{:<14.6e}".format(loc[1])
                outputData += "{:<+14.6e}".format(uloc)
                outputData += "{:<+14.6e}".format(vloc)
                outputData += "{:<+14.6e}".format(wloc)
                outputData += "{:<+14.6e}".format(phixloc)
                outputData += "{:<+14.6e}".format(phiyloc)
                outputData += "{:<+14.6e}".format(NMBloc[0])
                outputData += "{:<+14.6e}".format(NMBloc[1])
                outputData += "{:<+14.6e}".format(NMBloc[2])
                outputData += "{:<+14.6e}".format(NMBloc[3])
                outputData += "{:<+14.6e}".format(NMBloc[4])
                outputData += "{:<+14.6e}".format(NMBloc[5])
                outputData += "{:<+14.6e}".format(QBloc[0])
                outputData += "{:<+14.6e}\n".format(QBloc[1])
            
            outputData += "\n"

            if self.output_K == "yes":
                KData += f"LOADS ID: {lmb.loads.loadsID} PANEL ID: {lmb.panel.panelID}\n\n"
                KData += f"{lmb.KFull}\n\n"

            if self.output_P == "yes":
                PData += f"LOADS ID: {lmb.loads.loadsID} PANEL ID: {lmb.panel.panelID}\n\n"
                PData += f"{lmb.PFull}\n\n"

        with open(self.outputFile, 'w') as fh:
            fh.write(outputData)

        if self.output_K == "yes":
            with open("K.txt", 'w') as fh:
                fh.write(KData)

        if self.output_P == "yes":
            with open("P.txt", 'w') as fh:
                fh.write(PData)
        


import argparse
import configparser
import shutil
import sys
import traceback

from pathlib import Path

from lambend.__version__ import __version__
from lambend.expiry_wrapper import check_expiry
from lambend.lambendIO import LambendIO

def main():
    """Entry point of the command line application - lambend."""
    # check_expiry()

    parser = argparse.ArgumentParser(description='lambend : Analysis \
    of Anisotropic laminates Under Transverse Loads.')
    parser.add_argument("config", type=str, nargs='?', default=None, help="specify configuration file")
    parser.add_argument("-v", "--version", help="show version number", action='version', version='3pc-lambend ' + __version__)
    parser.add_argument("-s", "--sample", help="copy example files in the current directory", action="store_true")
    args = parser.parse_args()

    if args.sample:
        print('Copying sample file in current folder...')
        examples = Path(__file__).parent / 'examples'
        destination = Path(".") / 'examples'
        shutil.copytree(examples, destination, dirs_exist_ok=True)
        sys.exit()

    if args.config is None:
        sys.exit("No config file to run analysis. Try lambend -h")
    
    config = configparser.ConfigParser()
    config.read(args.config)

    materialFile = config["files"]["materials"]
    plyFile = config["files"]["plies"]
    laminateFile = config["files"]["laminates"]
    panelFile = config["files"]["panels"]
    loadsFile = config["files"]["loads"]
    outputFile = config["files"]["output"]
    number_of_terms = int(config["analysis"]["number_of_terms"])
    number_of_points_x = int(config["analysis"]["number_of_points_x"])
    number_of_points_y = int(config["analysis"]["number_of_points_y"])
    output_K = config["analysis"]["output_K"]
    output_K = output_K.lower()
    output_P = config["analysis"]["output_P"]
    output_P = output_P.lower()

    lmbIO = LambendIO(materialFile=materialFile, plyFile=plyFile,
    laminateFile=laminateFile, panelFile=panelFile, loadsFile=loadsFile,
    outputFile=outputFile, number_of_terms=number_of_terms,
    number_of_points_x=number_of_points_x, number_of_points_y=number_of_points_y,
    output_K=output_K, output_P=output_P)
    try:
        lmbIO._writeResults()
    except Exception:
        with open(outputFile, 'w') as fh:
            traceback.print_exc(limit=None, file=fh, chain=True)



from datetime import datetime
import sys

def check_expiry():
    expiry_date = datetime(2025, 12, 31)
    if datetime.now() > expiry_date:
        print("This application has expired.")
        sys.exit(1)


import copy
import numpy as np
from math import pi, sin, cos
from utils.utils import get_pairs, kronecker, delta1

class Lambend(object):

    def __init__(self, panel = None, loads = None, number_of_terms = None,
    number_of_points_x = None, number_of_points_y = None):
        """Initialize the Solver003 instance.
        
        Parameters
        ----------

        panel : panel.Panel (required)

        loads : loads.Loads (required)

        number_of_terms : int (required)

        """

        self.panel = panel
        self.loads = loads
        self.number_of_terms = int(number_of_terms)
        self.number_of_points_x = int(number_of_points_x)
        self.number_of_points_y = int(number_of_points_y)
        self._invalidate()

    @classmethod
    def fromDict(cls, **kwargs):
        """Initialize Solver003 from kwargs dictionary.
        
        Parameters
        ----------
        
        kwargs : dict

            {"panel": Panel, "loads": Loads, "number_of_terms": int}

        Returns
        -------

        cls : Solver003
        
        """
        panel = kwargs.get("panel")
        loads = kwargs.get("loads")
        number_of_terms = kwargs.get("number_of_terms")
        return cls(id, panel, loads, number_of_terms)

    def _invalidate(self):
        """Invalidate all calculation to force re calculation."""

        self._NT = None
        self._NC = None

        self._pairs = None
        self._K33 = None

        self._Pmn = None

        self._u = None
        self._v = None
        self._w = None
        self._phix = None
        self._phiy = None
        self._loc = None

        self._uloc = None
        self._vloc = None
        self._wloc = None
        self._phixloc = None
        self._phiyloc = None

        self._ex0Bloc = None
        self._ey0Bloc = None
        self._gxy0Bloc = None
        self._kx0Bloc = None
        self._ky0Bloc = None
        self._kxy0Bloc = None
        self._gxz0Bloc = None
        self._gyz0Bloc = None

        self._ek0Bloc = None
        self._gT0Bloc = None
        self._NMBloc = None
        self._QBloc = None

        self._coeff = None
        self._KFull = None
        self._PFull = None

    @property
    def NT(self):
        """Returns thermal inplane loads on the laminate."""

        if self._NT is not None:
            return self._NT
        
        self._NT = np.zeros(3)
        deltaT = self.loads.deltaT
        z = self.panel.laminate.z
        for i, ply in enumerate(self.panel.laminate.stacking_sequence, start=1):
            self._NT += (ply.Qbar @ ply.alphabar) * deltaT * (z[i]-z[i-1])
        return self._NT

    @property
    def NC(self):
        """Returns hygral inplane loads on the laminate."""

        if self._NC is not None:
            return self._NC
        
        self._NC = np.zeros(3)
        deltaC = self.loads.deltaC
        z = self.panel.laminate.z
        for i, ply in enumerate(self.panel.laminate.stacking_sequence, start=1):
            self._NC += (ply.Qbar @ ply.betabar) * deltaC * (z[i]-z[i-1])
        return self._NC

    @property
    def pairs(self):
        """List of pairs (i, j) for all i, j = 1, 2, ..., M, where
        M is number of terms in analysis.
        
        Returns
        -------
        
        pairs : [(1,1), (1,2), ...]
        
        """

        if self._pairs is not None:
            return self._pairs

        self._pairs = get_pairs(self.number_of_terms)
        return self._pairs
    
    @property
    def K33(self):
        """Part of K33 that depends on load
        
        Returns
        -------
        
        K33 : numpy.array

            Loads dependent part of K33 for all pairs

        """

        if self._K33 is not None:
            return self._K33

        self._K33 = [ [ self._calculate_K33(m, n, p, q)
        for p, q in self.pairs ] for m, n in self.pairs ]
        self._K33 = np.array(self._K33)
        return self._K33

    def _calculate_K33(self, m, n, p, q):
        L = self.panel.length
        W = self.panel.width
        a = m*pi/L
        b = n*pi/W
        Nxx = self.loads.Nxx+self.NT[0]+self.NC[0]
        Nyy = self.loads.Nyy+self.NT[1]+self.NC[1]
        Nxy = self.loads.Nxy+self.NT[2]+self.NC[2]
        val = 0.0
        val += 0.25*(Nxx*(a**2))*L*W*kronecker(m,p)*kronecker(n,q)
        val += 0.25*(Nyy*(b**2))*L*W*kronecker(m,p)*kronecker(n,q)
        val -= 8*Nxy*delta1(m,n,p,q)
        return val

    @property
    def Pmn(self):
        """Returns the loads vector
        
        Returns
        -------
        
        Pmn : numpy.array
        
        """

        if self._Pmn is not None:
            return self._Pmn

        self._Pmn = [self._calculate_Pmn(m ,n) for m, n in self.pairs]
        self._Pmn = np.array(self._Pmn)
        return self._Pmn

    @property
    def PFull(self):
        """Returns the full loads vector
        
        Returns
        -------
        
        PFull : numpy.array
        
        """

        if self._PFull is not None:
            return self._PFull

        dim = len(self.pairs)
        self._PFull = np.zeros(5*dim)
        self._PFull[2*dim:3*dim] = self.Pmn
        return self._PFull
    
    def _calculate_Pmn(self, m, n):
        """Calculated Pmn entries for a given m, n"""

        val = 0.0
        L = self.panel.length
        W = self.panel.width
        if self.loads.type == "sinusoidal pressure":
            p0 = self.loads.p0
            if m == 1 and n == 1:
                val = p0*L*W/4.0
        elif self.loads.type == "uniform pressure":
            p0 = self.loads.p0
            if m % 2 != 0 and n % 2 != 0:
                val = 4.0*p0*L*W/((pi**2)*m*n)
        elif self.loads.type == "point load":
            P = self.loads.P
            xCG = self.loads.xCG
            yCG = self.loads.yCG
            val = P*sin(m*pi*xCG/L)*sin(n*pi*yCG/W)
        elif self.loads.type == "partial pressure":
            p0 = self.loads.p0
            xCG = self.loads.xCG
            yCG = self.loads.yCG
            xLength = self.loads.xLength
            yLength = self.loads.yLength
            val = (4.0*p0*L*W/((pi**2)*m*n))*sin(m*pi*xCG/L)*sin(n*pi*yCG/W)*sin(m*pi*xLength/(2.0*L))*sin(n*pi*yLength/(2.0*W))
        elif self.loads.type == "hydrostatic pressure in x direction":
            p0 = self.loads.p0
            if n % 2 != 0:
                val = 2.0*p0*L*W*((-1)**(m+1))/((pi**2)*m*n)
        elif self.loads.type == "hydrostatic pressure in y direction":
            p0 = self.loads.p0
            if m % 2 != 0:
                val = 2.0*p0*L*W*((-1)**(n+1))/((pi**2)*m*n)
        elif self.loads.type == "triangular pressure in x direction":
            p0 = self.loads.p0
            if m % 2 != 0 and n % 2 != 0:
                val = 8.0*p0*L*W*((-1)**((m-1)/2))/((pi**3)*(m**2)*n)
        elif self.loads.type == "triangular pressure in y direction":
            p0 = self.loads.p0
            if m % 2 != 0 and n % 2 != 0:
                val = 8.0*p0*L*W*((-1)**((n-1)/2))/((pi**3)*(n**2)*m)
        return val

    @property
    def u(self):
        """Unknown coeffecient u_mn
        
        Returns
        -------

        u : numpy.array
        
        """

        if self._u is not None:
            return self._u

        dim = len(self.pairs)
        self._u = self.coeff[0:dim]
        return self._u

    @property
    def v(self):
        """Unknown coeffecient v_mn
        
        Returns
        -------

        v : numpy.array
        
        """

        if self._v is not None:
            return self._v

        dim = len(self.pairs)
        self._v = self.coeff[dim:2*dim]
        return self._u

    @property
    def w(self):
        """Unknown coeffecient w_mn
        
        [w] = K^-1 * Pmn

        Returns
        -------

        w : numpy.array
        
        """

        if self._w is not None:
            return self._w

        #self._w = np.linalg.inv(self.panel.K+self.K33) @ self.Pmn
        dim = len(self.pairs)
        self._w = self.coeff[2*dim:3*dim]
        return self._w

    @property
    def phix(self):
        """Unknown coeffecient phix_mn
        
        Returns
        -------

        phix : numpy.array
        
        """

        if self._phix is not None:
            return self._phix

        #self._w = np.linalg.inv(self.panel.K+self.K33) @ self.Pmn
        dim = len(self.pairs)
        self._phix = self.coeff[3*dim:4*dim]
        return self._phix

    @property
    def phiy(self):
        """Unknown coeffecient phiy_mn
        
        Returns
        -------

        phiy : numpy.array
        
        """

        if self._phiy is not None:
            return self._phiy

        #self._w = np.linalg.inv(self.panel.K+self.K33) @ self.Pmn
        dim = len(self.pairs)
        self._phiy = self.coeff[4*dim:5*dim]
        return self._phiy

    @property
    def loc(self):
        """Returns a list of location where w is computed
        
        Returns
        -------
        
        loc : [(0,0), (0,0.25), ...]
        
        """

        if self._loc is not None:
            return self._loc

        x = np.linspace(0, self.panel.length, num=self.number_of_points_x)
        y = np.linspace(0, self.panel.width, num=self.number_of_points_y)
        xv, yv = np.meshgrid(x, y)
        self._loc = [(xv[i,j], yv[i,j]) for i in range(self.number_of_points_y) for j in range(self.number_of_points_x)]
        return self._loc

    @property
    def uloc(self):
        """Returns a list of u computed at loc
        
        Returns
        -------
        
        uloc :[float, float, ...]
        
        """

        if self._uloc is not None:
            return self._uloc

        self._uloc = [self._computeUVal(x, y) for x, y in self.loc]
        return self._uloc

    def _computeUVal(self, x, y):

        L = self.panel.length
        W = self.panel.width
        uval = 0.0
        for index, pair in enumerate(self.pairs):
            m, n = pair
            uval += cos(m*pi*x/L)*sin(n*pi*y/W)*self.u[index]
        return uval

    @property
    def vloc(self):
        """Returns a list of v computed at loc
        
        Returns
        -------
        
        vloc :[float, float, ...]
        
        """

        if self._vloc is not None:
            return self._vloc

        self._vloc = [self._computeVVal(x, y) for x, y in self.loc]
        return self._vloc

    def _computeVVal(self, x, y):

        L = self.panel.length
        W = self.panel.width
        vval = 0.0
        for index, pair in enumerate(self.pairs):
            m, n = pair
            vval += sin(m*pi*x/L)*cos(n*pi*y/W)*self.v[index]
        return vval

    @property
    def wloc(self):
        """Returns a list of w computed at loc
        
        Returns
        -------
        
        wloc :[float, float, ...]
        
        """

        if self._wloc is not None:
            return self._wloc

        self._wloc = [self._computeWVal(x, y) for x, y in self.loc]
        return self._wloc

    def _computeWVal(self, x, y):

        L = self.panel.length
        W = self.panel.width
        wval = 0.0
        for index, pair in enumerate(self.pairs):
            m, n = pair
            wval += sin(m*pi*x/L)*sin(n*pi*y/W)*self.w[index]
        return wval

    @property
    def phixloc(self):
        """Returns a list of phix computed at loc
        
        Returns
        -------
        
        phixloc :[float, float, ...]
        
        """

        if self._phixloc is not None:
            return self._phixloc

        self._phixloc = [self._computePhixVal(x, y) for x, y in self.loc]
        return self._phixloc

    def _computePhixVal(self, x, y):

        L = self.panel.length
        W = self.panel.width
        phixval = 0.0
        for index, pair in enumerate(self.pairs):
            m, n = pair
            phixval += cos(m*pi*x/L)*sin(n*pi*y/W)*self.phix[index]
        return phixval

    @property
    def phiyloc(self):
        """Returns a list of phiy computed at loc
        
        Returns
        -------
        
        phiyloc :[float, float, ...]
        
        """

        if self._phiyloc is not None:
            return self._phiyloc

        self._phiyloc = [self._computePhiyVal(x, y) for x, y in self.loc]
        return self._phiyloc

    def _computePhiyVal(self, x, y):

        L = self.panel.length
        W = self.panel.width
        phiyval = 0.0
        for index, pair in enumerate(self.pairs):
            m, n = pair
            phiyval += sin(m*pi*x/L)*cos(n*pi*y/W)*self.phiy[index]
        return phiyval

    @property
    def ex0Bloc(self):
        """Returns a list of ex0 = du/dx computed at loc due to bending
        
        Returns
        -------
        
        ex0Bloc :[float, float, ...]
        
        """

        if self._ex0Bloc is not None:
            return self._ex0Bloc

        self._ex0Bloc = [self._computeEx0BVal(x, y) for x, y in self.loc]
        return self._ex0Bloc

    def _computeEx0BVal(self, x, y):

        L = self.panel.length
        W = self.panel.width
        ex0Bval = 0.0
        for index, pair in enumerate(self.pairs):
            m, n = pair
            ex0Bval += -(m*pi/L)*sin(m*pi*x/L)*sin(n*pi*y/W)*self.u[index]
        return ex0Bval
    
    @property
    def ey0Bloc(self):
        """Returns a list of ey0 = dv/dy computed at loc due to bending
        
        Returns
        -------
        
        ey0Bloc :[float, float, ...]
        
        """

        if self._ey0Bloc is not None:
            return self._ey0Bloc

        self._ey0Bloc = [self._computeEy0BVal(x, y) for x, y in self.loc]
        return self._ey0Bloc

    def _computeEy0BVal(self, x, y):

        L = self.panel.length
        W = self.panel.width
        ey0Bval = 0.0
        for index, pair in enumerate(self.pairs):
            m, n = pair
            ey0Bval += -(n*pi/W)*sin(m*pi*x/L)*sin(n*pi*y/W)*self.v[index]
        return ey0Bval

    @property
    def gxy0Bloc(self):
        """Returns a list of gxy0 = du/dy+dv/dx computed at loc due to bending
        
        Returns
        -------
        
        gxy0Bloc :[float, float, ...]
        
        """

        if self._gxy0Bloc is not None:
            return self._gxy0Bloc

        self._gxy0Bloc = [self._computeGxy0BVal(x, y) for x, y in self.loc]
        return self._gxy0Bloc

    def _computeGxy0BVal(self, x, y):

        L = self.panel.length
        W = self.panel.width
        gxy0Bval = 0.0
        for index, pair in enumerate(self.pairs):
            m, n = pair
            gxy0Bval += (((n*pi/W)*cos(m*pi*x/L)*cos(n*pi*y/W)*self.u[index])+
                         ((m*pi/L)*cos(m*pi*x/L)*cos(n*pi*y/W)*self.v[index]))
        return gxy0Bval

    @property
    def kx0Bloc(self):
        """Returns a list of kx0 = dphix/dx computed at loc due to bending
        
        Returns
        -------
        
        kx0Bloc :[float, float, ...]
        
        """

        if self._kx0Bloc is not None:
            return self._kx0Bloc

        self._kx0Bloc = [self._computeKx0BVal(x, y) for x, y in self.loc]
        return self._kx0Bloc

    def _computeKx0BVal(self, x, y):

        L = self.panel.length
        W = self.panel.width
        kx0Bval = 0.0
        for index, pair in enumerate(self.pairs):
            m, n = pair
            kx0Bval += -(m*pi/L)*sin(m*pi*x/L)*sin(n*pi*y/W)*self.phix[index]
        return kx0Bval
    
    @property
    def ky0Bloc(self):
        """Returns a list of ky0 = dphiy/dy computed at loc due to bending
        
        Returns
        -------
        
        ky0Bloc :[float, float, ...]
        
        """

        if self._ky0Bloc is not None:
            return self._ky0Bloc

        self._ky0Bloc = [self._computeKy0BVal(x, y) for x, y in self.loc]
        return self._ky0Bloc

    def _computeKy0BVal(self, x, y):

        L = self.panel.length
        W = self.panel.width
        ky0Bval = 0.0
        for index, pair in enumerate(self.pairs):
            m, n = pair
            ky0Bval += -(n*pi/W)*sin(m*pi*x/L)*sin(n*pi*y/W)*self.phiy[index]
        return ky0Bval

    @property
    def kxy0Bloc(self):
        """Returns a list of kxy0 = dphix/dy+dphiy/dx computed at loc due to bending
        
        Returns
        -------
        
        kxy0Bloc :[float, float, ...]
        
        """

        if self._kxy0Bloc is not None:
            return self._kxy0Bloc

        self._kxy0Bloc = [self._computeKxy0BVal(x, y) for x, y in self.loc]
        return self._kxy0Bloc

    def _computeKxy0BVal(self, x, y):

        L = self.panel.length
        W = self.panel.width
        kxy0Bval = 0.0
        for index, pair in enumerate(self.pairs):
            m, n = pair
            kxy0Bval += (((n*pi/W)*cos(m*pi*x/L)*cos(n*pi*y/W)*self.phix[index])+
                         ((m*pi/L)*cos(m*pi*x/L)*cos(n*pi*y/W)*self.phiy[index]))
        return kxy0Bval

    @property
    def gxz0Bloc(self):
        """Returns a list of gxz0 = phix+dw/dx computed at loc due to bending
        
        Returns
        -------
        
        gxz0Bloc :[float, float, ...]
        
        """

        if self._gxz0Bloc is not None:
            return self._gxz0Bloc

        self._gxz0Bloc = [self._computeGxz0BVal(x, y) for x, y in self.loc]
        return self._gxz0Bloc

    def _computeGxz0BVal(self, x, y):

        L = self.panel.length
        W = self.panel.width
        gxz0Bval = 0.0
        for index, pair in enumerate(self.pairs):
            m, n = pair
            gxz0Bval += ((cos(m*pi*x/L)*sin(n*pi*y/W)*self.phix[index])+
                         ((m*pi/L)*cos(m*pi*x/L)*sin(n*pi*y/W)*self.w[index]))
        return gxz0Bval

    @property
    def gyz0Bloc(self):
        """Returns a list of gyz0 = phiy+dw/dy computed at loc due to bending
        
        Returns
        -------
        
        gyz0Bloc :[float, float, ...]
        
        """

        if self._gyz0Bloc is not None:
            return self._gyz0Bloc

        self._gyz0Bloc = [self._computeGyz0BVal(x, y) for x, y in self.loc]
        return self._gyz0Bloc

    def _computeGyz0BVal(self, x, y):

        L = self.panel.length
        W = self.panel.width
        gyz0Bval = 0.0
        for index, pair in enumerate(self.pairs):
            m, n = pair
            gyz0Bval += ((sin(m*pi*x/L)*cos(n*pi*y/W)*self.phiy[index])+
                         ((n*pi/W)*sin(m*pi*x/L)*cos(n*pi*y/W)*self.w[index]))
        return gyz0Bval
    
    @property
    def KFull(self):
        """Returns KFull for Solver003
        
        Returns
        -------
        
        KFull: numpy.array
        
        """

        if self._KFull is not None:
            return self._KFull
        
        dim = len(self.pairs)
        _KFull = copy.deepcopy(self.panel.KFull)
        _KFull[2*dim:3*dim, 2*dim:3*dim] += self.K33
        self._KFull = _KFull
        return self._KFull

    @property
    def coeff(self):
        """Returns all the coeff using KFull
        
        Returns
        -------
        
        coeff: numpy.array
               [u, v, w, phi_x, phi_y]
        
        """

        if self._coeff is not None:
            return self._coeff

        self._coeff = np.linalg.solve(self.KFull, self.PFull)
        return self._coeff

    @property
    def ek0Bloc(self):
        """Returns (ex0, ey0, gxy0, kx0, ky0, kxy0) vec at every loc due to bending
        
        Returns
        -------
        
        ek0Bloc: [numpy.array, ...]
        
        """

        if self._ek0Bloc is not None:
            return self._ek0Bloc
        
        self._ek0Bloc = [np.array([self.ex0Bloc[index], self.ey0Bloc[index], 
        self.gxy0Bloc[index], self.kx0Bloc[index], self.ky0Bloc[index],
        self.kxy0Bloc[index]]) for index, pair in enumerate(self.loc)]
        return self._ek0Bloc

    @property
    def gT0Bloc(self):
        """Returns (gyz0, gxz0) vec at every loc due to bending
        
        Returns
        -------
        
        gT0loc: [numpy.array, ...]
        
        """

        if self._gT0Bloc is not None:
            return self._gT0Bloc
        
        self._gT0Bloc = [np.array([self.gyz0Bloc[index], self.gxz0Bloc[index]]) for index, pair in enumerate(self.loc)]
        return self._gT0Bloc

    @property
    def NMBloc(self):
        """Returns (Nxx, Nyy, Nxy, Mxx, Myy, Mxy) vec at every loc due to bending.
        
        Returns
        -------
        
        NMBloc: [numpy.array, ...]
        
        """

        if self._NMBloc is not None:
            return self._NMBloc

        self._NMBloc = [self.panel.laminate.ABD @ self.ek0Bloc[index] for index, pair in enumerate(self.loc)]

        return self._NMBloc

    @property
    def QBloc(self):
        """Returns (Qyz, Qxz) vec at every loc due to bending.
        
        Returns
        -------
        
        QBloc: [numpy.array, ...]
        
        """

        if self._QBloc is not None:
            return self._QBloc

        AT = np.array([[self.panel.laminate.A44, self.panel.laminate.A45],
        [self.panel.laminate.A45, self.panel.laminate.A55]])
        self._QBloc = [AT @ self.gT0Bloc[index] for index, pair in enumerate(self.loc)]

        return self._QBloc


import os
import re
from setuptools import setup, find_packages
from pathlib import Path

# Read the contents of your README file
this_directory = Path(__file__).parent
long_desc = (this_directory / "README.md").read_text()

def get_version():
    with open(os.path.join("lambend", "__version__.py")) as f:
        return re.search(r'__version__ = ["\'](.+?)["\']', f.read()).group(1)

setup(
    name='3pc-lambend',
    version=get_version(),
    description='Python library for analysis of shear deformable anisotropic laminates under transverse loading.',
    long_description=long_desc,
    long_description_content_type='text/markdown',
    author="Omprakash Seresta",
    author_email="oseresta@gmail.com",
    license="MIT",
    entry_points={
        'console_scripts': ['lambend = lambend.__main__:main']
    },
    packages=find_packages(),
    install_requires=[
        '3pc-material',
        '3pc-ply',
        '3pc-laminate',
        '3pc-panel',
        '3pc-loads',
        '3pc-utils',
        'numpy'
    ],
    keywords='laminate-bending composites transverse-loading shear-deformable anisotropic aerospace-engineering mechanics',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    project_urls={
        'Blog': 'https://3pcomposites.com/blogs-1/f/3pcsolver002',
        'Analysis Tools': 'https://3pcomposites.com/analysis-tools',
        'Documentation': 'https://drive.google.com/drive/folders/1jDb-P5BOgJ-moVU20B7D8oVeHjHLAHTm',
    },
    python_requires='>=3.6',
    include_package_data=True,
    zip_safe=False
)
